@@ -122,7 +122,7 @@ static  int      tw_collect		(COMMAND, int, struct Strbuf *,
 static	Char 	 tw_suffix		(int, struct Strbuf *,const Char *,
 					 Char *);
 static	void 	 tw_fixword		(int, struct Strbuf *, Char *, Char *);
-static	void	 tw_list_items		(int, int, int);
+static	void	 tw_list_items		(const Char *, int, int, int);
 static 	void	 add_scroll_tab		(Char *);
 static 	void 	 choose_scroll_tab	(struct Strbuf *, int);
 static	void	 free_scroll_tab	(void);
@@ -234,13 +234,13 @@ tenematch(Char *inputline, int num_read, COMMAND command)
 	const Char *p;
 
 	xprintf(CGETS(30, 1, "starting_a_command %d\n"), looking);
-	xprintf("\ncmd_start:%S:\n", qline.s + cmd_start);
-	xprintf("qline:%S:\n", qline.s);
+	xprintf("\ncmd_start:%" TCSH_S ":\n", qline.s + cmd_start);
+	xprintf("qline:%" TCSH_S ":\n", qline.s);
 	xprintf("qline:");
 	for (p = qline.s; *p; p++)
 	    xprintf("%c", *p & QUOTE ? '-' : ' ');
 	xprintf(":\n");
-	xprintf("word:%S:\n", qline.s + word);
+	xprintf("word:%" TCSH_S ":\n", qline.s + word);
 	xprintf("word:");
 	for (p = qline.s + word; *p; p++)
 	    xprintf("%c", *p & QUOTE ? '-' : ' ');
@@ -260,7 +260,7 @@ tenematch(Char *inputline, int num_read, COMMAND command)
 	looking = tw_complete(qline.s + cmd_start, &p, &pat, looking, &suf);
 	wordp = p - qline.s;
 #ifdef TDEBUG
-	xprintf(CGETS(30, 3, "complete %d %S\n"), looking, pat);
+	xprintf(CGETS(30, 3, "complete %d %" TCSH_S "\n"), looking, pat);
 #endif
     }
 
@@ -393,7 +393,7 @@ tenematch(Char *inputline, int num_read, COMMAND command)
 	    cleanup_push(ptr, blk_cleanup);
 	if (count > 0) {
 	    if (command == GLOB)
-		print_by_column(STRNULL, ptr, count, 0);
+		print_by_column(STRNULL, ptr, count, FALSE);
 	    else {
 		DeleteBack(str_end - word_start);/* get rid of old word */
 		for (i = 0; i < count; i++)
@@ -981,7 +981,7 @@ tw_collect_items(COMMAND command, int looking, struct Strbuf *exp_dir,
 	    tw_next_entry[looking](&item, exp_dir, &flags) != 0)) {
 	Strbuf_terminate(&item);
 #ifdef TDEBUG
-	xprintf("item = %S\n", item.s);
+	xprintf("item = %" TCSH_S "\n", item.s);
 #endif
 	switch (looking) {
 	case TW_FILE:
@@ -1135,13 +1135,13 @@ tw_collect_items(COMMAND command, int looking, struct Strbuf *exp_dir,
 		case TW_COMMAND:
 		    if (!(dir_ok && exec_check))
 			break;
-		    if (filetype(exp_dir->s, item.s) == '/')
+		    if (get_filetype(exp_dir->s, item.s, TRUE).suffix == '/')
 			Strbuf_append1(&buf, '/');
 		    break;
 
 		case TW_FILE:
 		case TW_DIRECTORY:
-		    Strbuf_append1(&buf, filetype(exp_dir->s, item.s));
+		    Strbuf_append1(&buf, get_filetype(exp_dir->s, item.s, TRUE).suffix);
 		    break;
 
 		default:
@@ -1152,7 +1152,7 @@ tw_collect_items(COMMAND command, int looking, struct Strbuf *exp_dir,
                      || looking == TW_LOGNAME) && tw_item_find(buf.s))
 		    break;
 		else {
-		    /* maximum length 1 (NULL) + 1 (~ or $) + 1 (filetype) */
+		    /* maximum length 1 (NULL) + 1 (~ or $) + 1 (suffix) */
 		    tw_item_add(&buf);
 		    if (command == LIST)
 			numitems++;
@@ -1197,7 +1197,7 @@ tw_collect_items(COMMAND command, int looking, struct Strbuf *exp_dir,
 	    break;
 	}
 #ifdef TDEBUG
-	xprintf("done item = %S\n", item.s);
+	xprintf("done item = %" TCSH_S "\n", item.s);
 #endif
     }
     cleanup_until(&item);
@@ -1349,7 +1349,7 @@ tw_collect(COMMAND command, int looking, struct Strbuf *exp_dir,
     jmp_buf_t osetexit;
 
 #ifdef TDEBUG
-    xprintf("target = %S\n", target);
+    xprintf("target = %" TCSH_S "\n", target);
 #endif
     ni = 0;
     getexit(osetexit);
@@ -1399,7 +1399,7 @@ tw_collect(COMMAND command, int looking, struct Strbuf *exp_dir,
  *	(by default interpreted as 'items', for backwards compatibility)
  */
 static void
-tw_list_items(int looking, int numitems, int list_max)
+tw_list_items(const Char * dir, int looking, int numitems, int list_max)
 {
     Char *ptr;
     int max_items = 0;
@@ -1477,7 +1477,7 @@ tw_list_items(int looking, int numitems, int list_max)
     if (looking != TW_SIGNAL)
 	qsort(tw_item_get(), numitems, sizeof(Char *), fcompare);
     if (looking != TW_JOB)
-	print_by_column(STRNULL, tw_item_get(), numitems, TRUE);
+	print_by_column(dir, tw_item_get(), numitems, TRUE);
     else {
 	/*
 	 * print one item on every line because jobs can have spaces
@@ -1487,7 +1487,7 @@ tw_list_items(int looking, int numitems, int list_max)
 	Char **w = tw_item_get();
 
 	for (i = 0; i < numitems; i++) {
-	    xprintf("%S", w[i]);
+	    xprintf("%" TCSH_S "", w[i]);
 	    if (Tty_raw_mode)
 		xputchar('\r');
 	    xputchar('\n');
@@ -1599,7 +1599,7 @@ t_search(struct Strbuf *word, COMMAND command, int looking, int list_max,
 
     case TW_EXPLAIN:
 	if (command == LIST && pat != NULL) {
-	    xprintf("%S", pat);
+	    xprintf("%" TCSH_S "", pat);
 	    if (Tty_raw_mode)
 		xputchar('\r');
 	    xputchar('\n');
@@ -1695,7 +1695,7 @@ t_search(struct Strbuf *word, COMMAND command, int looking, int list_max,
 	if ((dir_fd = opendir(short2str(exp_dir.s))) == NULL) {
  	    if (command == RECOGNIZE)
  		xprintf("\n");
- 	    xprintf("%S: %s", exp_dir.s, strerror(errno));
+	    xprintf("%" TCSH_S ": %s", exp_dir.s, strerror(errno));
  	    if (command != RECOGNIZE)
  		xprintf("\n");
  	    NeedsRedraw = 1;
@@ -1805,7 +1805,7 @@ t_search(struct Strbuf *word, COMMAND command, int looking, int list_max,
 	break;
 
     case LIST:
-	tw_list_items(looking, numitems, list_max);
+	tw_list_items(exp_dir.s, looking, numitems, list_max);
 	tw_item_free();
 	break;
 
@@ -1956,7 +1956,7 @@ expand_dir(const Char *dir, struct Strbuf *edir, DIR **dfd, COMMAND cmd)
 	 * From: Amos Shapira <amoss@cs.huji.ac.il>
 	 * Print a better message when completion fails
 	 */
-	xprintf("\n%S %s\n", edir->len ? edir->s : (tdir ? tdir : dir),
+	xprintf("\n%" TCSH_S " %s\n", edir->len ? edir->s : (tdir ? tdir : dir),
 		(errno == ENOTDIR ? CGETS(30, 10, "not a directory") :
 		(errno == ENOENT ? CGETS(30, 11, "not found") :
 		 CGETS(30, 12, "unreadable"))));
@@ -2009,13 +2009,20 @@ nostat(const Char *dir)
 } /* end nostat */
 
 
-/* filetype():
- *	Return a character that signifies a filetype
- *	symbology from 4.3 ls command.
+/* get_filetype():
+ *	Return a struct filetype based on the file mode.
  */
-Char
-filetype(const Char *dir, const Char *file)
+struct filetype
+get_filetype(const Char *dir, const Char *file, int use_lstat)
 {
+#ifdef COLOR_LS_F
+#define RETURN_FILETYPE(suffix,color) do { \
+    struct filetype r={suffix,color}; return r; } while (0)
+#else
+#define RETURN_FILETYPE(suffix,color) do { \
+    struct filetype r={suffix}; return r; } while (0)
+#endif /* COLOR_LS_F */
+
     Char *path;
     char   *ptr;
     struct stat statb;
@@ -2027,34 +2034,39 @@ filetype(const Char *dir, const Char *file)
     ptr = short2str(path);
     xfree(path);
 
-    if (lstat(ptr, &statb) == -1)
-	goto out;
+    if (use_lstat) {
+	if (lstat(ptr, &statb) == -1)
+	    goto out;
+    } else {
+	if (stat(ptr, &statb) == -1)
+	    goto out;
+    }
 
 #ifdef S_ISLNK
     if (S_ISLNK(statb.st_mode)) {	/* Symbolic link */
 	if (adrof(STRlistlinks)) {
 	    if (stat(ptr, &statb) == -1)
-		return '&';
+		RETURN_FILETYPE('&', CV_ORPHAN);
 	    else if (S_ISDIR(statb.st_mode))
-		return '>';
+		RETURN_FILETYPE('>', CV_LNK_DIR);
 	    else
-		return '@';
+		RETURN_FILETYPE('@', CV_LNK);
 	}
 	else
-	    return '@';
+	    RETURN_FILETYPE('@', CV_LNK);
     }
 #endif
 #ifdef S_ISSOCK
     if (S_ISSOCK(statb.st_mode))	/* Socket */
-	return '=';
+	RETURN_FILETYPE('=', CV_SOCK);
 #endif
 #ifdef S_ISFIFO
     if (S_ISFIFO(statb.st_mode)) /* Named Pipe */
-	return '|';
+	RETURN_FILETYPE('|', CV_FIFO);
 #endif
 #ifdef S_ISHIDDEN
     if (S_ISHIDDEN(statb.st_mode)) /* Hidden Directory [aix] */
-	return '+';
+	RETURN_FILETYPE('+', CV_HIDDEN);
 #endif
 #ifdef S_ISCDF
     {
@@ -2064,32 +2076,52 @@ filetype(const Char *dir, const Char *file)
 	p2 = strspl(ptr, "+");	/* Must append a '+' and re-stat(). */
 	if ((stat(p2, &hpstatb) != -1) && S_ISCDF(hpstatb.st_mode)) {
 	    xfree(p2);
-	    return '+';	/* Context Dependent Files [hpux] */
+	    RETURN_FILETYPE('+', CV_CDF);	/* Context Dependent Files [hpux] */
 	}
 	xfree(p2);
     }
 #endif
 #ifdef S_ISNWK
     if (S_ISNWK(statb.st_mode)) /* Network Special [hpux] */
-	return ':';
+	RETURN_FILETYPE(':', CV_NWK);
 #endif
 #ifdef S_ISCHR
     if (S_ISCHR(statb.st_mode))	/* char device */
-	return '%';
+	RETURN_FILETYPE('%', CV_CHR);
 #endif
 #ifdef S_ISBLK
     if (S_ISBLK(statb.st_mode))	/* block device */
-	return '#';
+	RETURN_FILETYPE('#', CV_BLK);
 #endif
 #ifdef S_ISDIR
-    if (S_ISDIR(statb.st_mode))	/* normal Directory */
-	return '/';
+    if (S_ISDIR(statb.st_mode)) {
+	if ((statb.st_mode & S_ISVTX) && (statb.st_mode & S_IWOTH))
+	    RETURN_FILETYPE('/', CV_DIR_TW);	/* +t o+w directory */
+	if (statb.st_mode & S_IWOTH)
+	    RETURN_FILETYPE('/', CV_DIR_OW);	/* o+w directory */
+	if (statb.st_mode & S_ISVTX)
+	    RETURN_FILETYPE('/', CV_DIR_ST);	/* +t directory */
+	RETURN_FILETYPE('/', CV_DIR);		/* normal directory */
+    }
 #endif
-    if (statb.st_mode & (S_IXUSR|S_IXGRP|S_IXOTH))
-	return '*';
+    if (S_ISREG(statb.st_mode)) {
+	const int isexec = statb.st_mode & (S_IXUSR|S_IXGRP|S_IXOTH);
+	const Char suf = isexec ? '*' : ' ';
+	if (statb.st_mode & S_ISUID)
+	    RETURN_FILETYPE(suf, CV_SUID);	/* setuid */
+	if (statb.st_mode & S_ISGID)
+	    RETURN_FILETYPE(suf, CV_SGID);	/* setgid */
+	if (isexec)
+	    RETURN_FILETYPE('*', CV_EXE);
+	if (statb.st_nlink > 1)
+	    RETURN_FILETYPE(' ', CV_HARD);
+	RETURN_FILETYPE(' ', CV_FILE);
+    }
 out:
-    return ' ';
-} /* end filetype */
+    RETURN_FILETYPE(' ', CV_FILE);
+
+#undef RETURN_FILETYPE
+} /* end get_filetype */
 
 
 /* isadirectory():
@@ -2154,7 +2186,7 @@ find_rows(Char *items[], int count, int no_file_suffix)
  *
  */
 void
-print_by_column(Char *dir, Char *items[], int count, int no_file_suffix)
+print_by_column(const Char *dir, Char *items[], int count, int no_file_suffix)
 {
     int i, r, c, columns, rows;
     size_t w;
@@ -2189,25 +2221,29 @@ print_by_column(Char *dir, Char *items[], int count, int no_file_suffix)
 
 #ifdef COLOR_LS_F
 		if (no_file_suffix) {
-		    /* Print the command name */
+		    /* Print the non-file name */
 		    Char f = items[i][w - 1];
 		    items[i][w - 1] = 0;
-		    print_with_color(items[i], w - 1, f);
+		    struct filetype ft = get_filetype(dir, items[i], TRUE);
+		    ft.suffix = f;
+		    print_with_color(dir, items[i], w - 1, ft);
 		    items[i][w - 1] = f;
 		}
 		else {
-		    /* Print filename followed by '/' or '*' or ' ' */
-		    print_with_color(items[i], w, filetype(dir, items[i]));
+		    /* Print filename in color based on filetype */
+		    print_with_color(dir, items[i], w,
+			get_filetype(dir, items[i], TRUE));
 		    wx++;
 		}
 #else /* ifndef COLOR_LS_F */
 		if (no_file_suffix) {
-		    /* Print the command name */
-		    xprintf("%S", items[i]);
+		    /* Print the non-file name */
+		    xprintf("%" TCSH_S, items[i]);
 		}
 		else {
 		    /* Print filename followed by '/' or '*' or ' ' */
-		    xprintf("%-S%c", items[i], filetype(dir, items[i]));
+		    xprintf("%-" TCSH_S "%c", items[i],
+			get_filetype(dir, items[i], TRUE).suffix);
 		    wx++;
 		}
 #endif /* COLOR_LS_F */
